﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;
using ucore_assembler_library;
using System.Runtime.InteropServices;
using Microsoft.Win32.SafeHandles;
using raw_device_io;
using xcore_tools;

namespace RAWEdit
{
    public partial class RAWEditMainForm : Form, IPrint
    {
        raw_device_access m_raw_device_access = new raw_device_access();
        Buffer m_buffer = new Buffer();
        UInt64 m_rawReadCounter = 0;

        string m_readDestinationFileName = "";
        string m_loadFileIntoBufferFileName = "";

        bool m_enableCloseDiaglog = true;
      

        public RAWEditMainForm()
        {
            InitializeComponent();
            buttonRefreshDeviceList_Click(null, null);       
        }

        public bool enableCloseDialog
        {
            get { return m_enableCloseDiaglog; }
            set { m_enableCloseDiaglog = value; }
        }

        private void buttonRefreshDeviceList_Click(object sender, EventArgs e)
        {
            refreshDeviceList(); 
            comboBoxDevices.Items.Clear();

            foreach (DriveInfo di in m_deviceInfos)
            {
                string idEnt = string.Format("{0} {1,-16} Type {2,-12} Size = {3:0.000}MB", di.Name, di.IsReady ? di.VolumeLabel : "NA",  di.DriveType.ToString() , di.IsReady ? ((double)di.TotalSize) / 1048576.0 : 0);
                comboBoxDevices.Items.Add(idEnt);
            }

            if (m_deviceInfos.Count > 0)
            {
                comboBoxDevices.SelectedIndex = 0;
            }
              
        }



        List<DriveInfo> m_deviceInfos = null;
        DriveInfo m_currentDevice = null;

        void refreshDeviceList()
        {
            m_deviceInfos = m_raw_device_access.getDevices(DriveType.Removable);    // only removeble devices                            
        }

        #region IPrint
        public void print(string txt)
        {
            debugBox.AppendText(txt);
        }

        public void dprint(string txt)
        {
            debugBox.AppendText(txt);
        }
        #endregion

        private void comboBoxDevices_SelectedIndexChanged(object sender, EventArgs e)
        {
            m_currentDevice = m_deviceInfos.ElementAt(comboBoxDevices.SelectedIndex);
        }

        
      
        private void buttonTestRead_Click(object sender, EventArgs e)
        {
            if (m_currentDevice != null)
            {

                try
                {
                    int readStart = (int)numericUpDownStartRead.Value;
                    int readLength = (int)numericUpDownLengthRead.Value;
                    string deviceName = m_currentDevice.Name;

                    byte[] rData = m_raw_device_access.readFromDevice(deviceName, readStart, readLength);

                    // print out

                    for (int i = 0; i < rData.Length; i++)
                    {
                        print(string.Format("{0:X2}", rData[i]));
                        if (i % 40 == 0)
                        {
                            print("\n");
                        }
                    }
                    print("\n");

                    for (int i = 0; i < rData.Length; i++)
                    {
                        char x = (char)rData[i];
                        if (x < 32 || x > 126)
                        {
                            x = ' ';
                        }
                        print(string.Format("{0}", x));
                        if (i % 40 == 0)
                        {
                            print("\n");
                        }
                    }
                    print("\n");

                }
                catch (Exception ex)
                {
                   print("Read access error, reason = " + ex.Message + "\n");
                }
            }       
        }

        public void refreshBufferView()
        { 
            List<BufferEntry> bEntrys = m_buffer.entrys;

            if (bEntrys != null && bEntrys.Count > 0)
            {
                dataGridViewBuffer.RowCount = bEntrys.Count;

                for (int i = 0; i < bEntrys.Count; i++)
                {
                    BufferEntry bEntry = bEntrys[i];
                    dataGridViewBuffer[0, i].Value = bEntry.name;
                    dataGridViewBuffer[1, i].Value = bEntry.data.Length;
                }
            }

        }


        private void buttonReadToFile_Click(object sender, EventArgs e)
        {
            int readStart = (int)numericUpDownStartRead.Value;
            int readLength = (int)numericUpDownLengthRead.Value;

            if (m_currentDevice == null)    // no device selected
            {
                return;
            }

            string deviceName = m_currentDevice.Name;

            try
            {
                print("Read ... \n");
                byte[] rData = m_raw_device_access.readFromDevice(deviceName, readStart, readLength);  
              
                m_buffer.add(new BufferEntry("RAW Read (" + m_rawReadCounter + ")",rData));
                m_rawReadCounter++;

                refreshBufferView();

            }
            catch (Exception)
            {
                print("Error while selecting/open file or dir\n");
            }
        }

   
        private void buttonLoadFileIntoBuffer_Click(object sender, EventArgs e)
        {
            try
            {
                OpenFileDialog openFileDialog = new OpenFileDialog();

                openFileDialog.FileName = m_loadFileIntoBufferFileName;


                openFileDialog.InitialDirectory = System.IO.Directory.GetCurrentDirectory();
                openFileDialog.Filter = "All files (*.*)|*.*";
                openFileDialog.FilterIndex = 2;
                openFileDialog.RestoreDirectory = true;
                openFileDialog.Multiselect = true;
               
                if (openFileDialog.ShowDialog() == DialogResult.OK)
                {
                    int fileCount = openFileDialog.FileNames.Count();

                    string[] fileNames = new string[fileCount];
                    for(int i=0;i<fileCount;i++)
                    {
                        fileNames[i] = openFileDialog.FileNames[i]; 
                    }

                    addFilesToBuffer(fileNames);
                    refreshBufferView();
                }
            }
            catch (Exception ex)
            {
                print("Error: " + ex.Message + "\n"); 
            }
        }

        public void clearFileBuffer()
        {
            m_buffer.clear();
            refreshBufferView();
        }

        public void addFilesToBuffer(string[] fileNames)
        {
            foreach (string fileName in fileNames)
            {
                m_loadFileIntoBufferFileName = fileName;

                print("Read File ... \n");

                try
                { 
                    using (FileStream fs = File.Open(m_loadFileIntoBufferFileName, FileMode.Open, FileAccess.Read, FileShare.None))
                    {
                        long fileLength = fs.Length;
                        long fileLengthBlocks = fileLength / 512;
                        if (fileLengthBlocks * 512 < fileLength)
                        {
                            fileLengthBlocks++;
                        }

                        print("File length = " + fileLength + " norm to " + fileLengthBlocks + " blocks (" + (fileLengthBlocks * 512) + " bytes)\n");

                        byte[] rData = new byte[fileLengthBlocks * 512];
                        for (int i = 0; i < (fileLengthBlocks * 512); i++)
                        {
                            rData[i] = 0;
                        }

                        fs.Read(rData, 0, (int)fileLength);

                        m_buffer.add(new BufferEntry("File Read '" + fileName + "' (" + m_rawReadCounter + ")", rData));
                        m_rawReadCounter++;

                        fs.Close();
                    }
                }
                catch(Exception ex)
                {
                    print("Error while loading file '" + fileName + "' Detailed = '" + ex.Message +"'\n");
                }
                
            }
        }

        private void buttonWriteFromBuffer_Click(object sender, EventArgs e)
        {
            int startAddress = (int)numericUpDownWriteStart.Value;

            List<BufferEntry> entrys = m_buffer.entrys;

            if (m_currentDevice == null)    // no device selected
            {
                return;
            }

            string deviceName = m_currentDevice.Name;

            try
            {
                foreach (BufferEntry entry in entrys)
                {
                    int entryLength = entry.data.Length;

                    print(string.Format("Write to device {0} at offset {1} {2} bytes (name = '{3}')\n", deviceName, startAddress, entryLength, entry.name));
                    m_raw_device_access.writeToDevice(deviceName, startAddress, entryLength, entry.data, false);

                    startAddress += entryLength;
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(this, "Error: " + ex.Message, "Something happen ...", MessageBoxButtons.OK);
            }
        }

        private void buttonClearBuffer_Click(object sender, EventArgs e)
        {
            clearFileBuffer();
        }

        private void buttonWriteBuffertoFile_Click(object sender, EventArgs e)
        {
            try
            {
                SaveFileDialog saveFileDialog = new SaveFileDialog();

                saveFileDialog.InitialDirectory = System.IO.Directory.GetCurrentDirectory();
                saveFileDialog.Filter = "All files (*.*)|*.*";
                saveFileDialog.FilterIndex = 2;
                saveFileDialog.RestoreDirectory = true;

                if (saveFileDialog.ShowDialog() == DialogResult.OK)
                {
                    string fileName = saveFileDialog.FileName;
                    using (FileStream fs = File.Open(fileName, FileMode.Create, FileAccess.Write, FileShare.None))
                    {
                        List<BufferEntry> entrys = m_buffer.entrys;

                        foreach (BufferEntry entry in entrys)
                        {
                            fs.Write(entry.data, 0, entry.data.Length);
                        }

                        print("Write finish ... \n");

                        fs.Close();
                    }
                }
            }
            catch (Exception ex)
            {
                print("Error: " + ex.Message + "\n");
            }
                
        }

        private void RAWEditMainForm_Shown(object sender, EventArgs e)
        {
            if (MessageBox.Show("NOTE! This tool can access storage devices in RAW mode!\nIf you dont know what you do, you can damage your data!\nDo you want to continue?", "WARNING", MessageBoxButtons.YesNo, MessageBoxIcon.Stop) == System.Windows.Forms.DialogResult.No)
            {
                Close();
            }
        }

        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Close();
        }

        private void button1_Click(object sender, EventArgs e)
        {
            string deviceName = m_currentDevice.Name;

            int clearLength = (int)numericUpDownClearLength.Value;

            byte []initData = new byte[clearLength];
            for(int i=0;i<clearLength;i++)
            {
                initData[i] = 0xff;
            }
            print("Clear device (" + deviceName + ") clear length = " + clearLength + "\n");
            m_raw_device_access.writeToDevice(deviceName, 0, clearLength, initData, false);

        }
    }
}
