﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using xcore_tools;
using xcore_tools.Common;

namespace hcore_assembler_library.Opcode
{
    public class Programm : IPrint
    {
        IPrint m_iprint;
        List<HOpcode> m_opcodes;
        public Labels labels
        {
            get;
            set;
        }

        public Programm(IPrint iprint)
        {
            m_iprint = iprint;
            labels = new Labels();
            m_opcodes = new List<HOpcode>();
            
        }

        public void addOpcode(HOpcode o)
        {
            m_opcodes.Add(o);
        }

        public void preresolve()
        {
            uint pc = 0xdf100000;

            foreach (HOpcode o in m_opcodes)
            {
                pc = o.preresolve(pc);             
            }
        }

        public void resolve()
        {
            foreach (HOpcode o in m_opcodes)
            {
                o.resolve();
            }
        }

        public List<uint> assembleAndGetData()
        {
            List<uint> r = new List<uint>();

            foreach (HOpcode o in m_opcodes)
            {
                o.assemble();
                bool hasData;
                uint data = o.getData(out hasData);
                if (hasData)
                {
                    r.Add(data);
                }
            }
            return r;
        }
           

        public void trace()
        {
            foreach (HOpcode o in m_opcodes)
            {
                o.assemble();
                bool hasData;
                uint data = o.getData(out hasData);
                if (hasData)
                {
                    print(string.Format("{0:X8}:{1:X8} => {2}\n",o.getPc(),data,o.trace()));                   
                }
            }
        }

        public void print(string txt)
        {
            m_iprint.print(txt);
        }

        public void dprint(string txt)
        {
            m_iprint.dprint(txt);
        }

        public void addDefines(Dictionary<string, string> defines)
        {
            foreach (KeyValuePair<string, string> kv in defines)
            {
                string name = kv.Key;
                try
                {
                    uint value = (uint)Tools.extractInteger(kv.Value);
                    labels.addLabel(name, value);
                }
                catch (Exception)
                {
                    // only add defines with integer values (proccess via exception is slow, so fix me)
                }
            }
        }

        public void check()
        {
            List<HOpcode> prg = new List<HOpcode>();

            foreach (HOpcode o in m_opcodes)
            {
                o.assemble();
                bool hasData;
                uint data = o.getData(out hasData);
                if (hasData)
                {
                    prg.Add(o);
                }
            }

            int length = prg.Count();

            for (int i = 0; i < length - 4; i++)
            {
                HOpcode o0 = prg[i + 0];
                HOpcode o1 = prg[i + 1];
                HOpcode o2 = prg[i + 2];
                HOpcode o3 = prg[i + 3];

                // test RAW hazard (in order)

                if (o0.useCAsReg && o0.isRCWE && !o0.canLockReg)   // o0 use dest register and will write to them and is not out of order with reg lock
                {
                    int tR = o0.RC;
                    xcore_tools.Data.Line o0Line = o0.opcodeInfo.line;
                    if (o0Line == null)
                    {
                        o0Line = new xcore_tools.Data.Line("unknown", "unknown", 0);
                    }

                    if (o1.useAAsReg && o1.RA == tR)
                    {
                        print(string.Format("RAW Hazard +1 Cycle RA read r{0} write r{1} at Line {2} in File {3}\n", o1.RA, tR, o0Line.line, o0Line.filename));
                    }
                    if (o1.useBAsReg && o1.RB == tR)
                    {
                        print(string.Format("RAW Hazard +1 Cycle RB read r{0} write r{1} at Line {2} in File {3}\n", o1.RB, tR, o0Line.line, o0Line.filename));
                    }

                    if (o2.useAAsReg && o2.RA == tR)
                    {
                        print(string.Format("RAW Hazard +2 Cycle RA read r{0} write r{1} at Line {2} in File {3}\n", o2.RA, tR, o0Line.line, o0Line.filename));
                    }
                    if (o2.useBAsReg && o2.RB == tR)
                    {
                        print(string.Format("RAW Hazard +2 Cycle RB read r{0} write r{1} at Line {2} in File {3}\n", o2.RB, tR, o0Line.line, o0Line.filename));
                    }
                }

                // test WAL Deadlock (out of order)
                if (o0.useCAsReg && o0.isRCWE && o0.canLockReg)   // o0 use dest register and will write to them and is  out of order with reg lock
                {
                    int tR = o0.RC;
                    xcore_tools.Data.Line o0Line = o0.opcodeInfo.line;
                    if (o0Line == null)
                    {
                        o0Line = new xcore_tools.Data.Line("unknown", "unknown", 0);
                    }

                    if (o1.useAAsReg && o1.RA == tR)
                    {
                        print(string.Format("Register Lock Hazard (WAL Wait and Lock) +1 Cycle RA read r{0} write r{1} at Line {2} in File {3}\n", o1.RA, tR, o0Line.line, o0Line.filename));
                    }
                    if (o1.useBAsReg && o1.RB == tR)
                    {
                        print(string.Format("Register Lock Hazard (WAL Wait and Lock)  +1 Cycle RB read r{0} write r{1} at Line {2} in File {3}\n", o1.RB, tR, o0Line.line, o0Line.filename));
                    }

                   /* if (o2.useAAsReg && o2.RA == tR)
                    {
                        print(string.Format("Register Lock Hazard (WAL Wait and Lock) +2 Cycle RA read r{0} write r{1} at Line {2} in File {3}\n", o2.RA, tR, o0Line.line, o0Line.filename));
                    }
                    if (o2.useBAsReg && o2.RB == tR)
                    {
                        print(string.Format("Register Lock Hazard (WAL Wait and Lock)  +2 Cycle RB read r{0} write r{1} at Line {2} in File {3}\n", o2.RB, tR, o0Line.line, o0Line.filename));
                    }*/
                }
            }

        }
    }
}
