﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using ucore_assembler_library;
using xcore_tools;
using xcore_tools.Common;

namespace ucore_assembler_console
{
    class Program
    {
        static void Main(string[] args)
        {
            string sourceFileName = "";
            string destinationFileName = "";
            bool debugLog = false;
            bool waitKey = false;
            uint stlength = 16384;
            bool outFileEnabled = false;
            uint warningLevel = 2;
            bool enableTrace = false;
            bool ignoreLength = false;
            List<Assembler.storeTypes_t> m_storeTypes = new List<Assembler.storeTypes_t>();
            

            int argCount = args.Length;

            if (argCount == 0)
            {
                print("Usage: ucore_assembler_console -s sourceFile [-o destFile] [-wkey][-dlog][...]\n");
                print("\n");
                print("Description:   -s          source file\n");
                print("               -o          output file (will extended with .mif or .raw)\n");
                print("               -wkey       wait for key after programm end\n");
                print("               -wlevel     set warning level (2 is default)\n");
                print("               -trace      enable trace\n");
                print("               -dlog       enable debug log\n");
                print("               -smif       store as MIF\n");
                print("               -sraw       store as raw\n");
                print("               -sdebug     store as debug (emulator)\n");
                print("               -stlength   set fixed store address count (default 2048)\n");
                print("               -stnotfixed set no fixed storesize\n");

                return;
            }

            for (uint i = 0; i < argCount; i++)
            {
                switch (args[i])
                {

                    case "-smif":
                        m_storeTypes.Add(Assembler.storeTypes_t.st_mif);
                        break;
                    case "-sraw":
                        m_storeTypes.Add(Assembler.storeTypes_t.st_raw);
                        break;
                    case "-sdebug":
                        m_storeTypes.Add(Assembler.storeTypes_t.st_debug);
                        break;

                    case "-wkey":
                        waitKey = true;
                        break;

                    case "-wlevel":
                        if (i + 1 >= argCount)
                        {
                            print("error: no level after -wlevel\n");
                            return;
                        }

                        warningLevel = Convert.ToUInt32(args[i + 1]);
                       
                        break;                 

                    case "-trace":
                        enableTrace = true;
                        break;

                    case "-dlog":
                        debugLog = true;
                        break;

                    case "-s":

                        if (i + 1 >= argCount)
                        {
                            print("error: no source file after -s\n");
                            return;
                        }

                        sourceFileName = args[i + 1];

                        break;

                    case "-stlength":

                        if (i + 1 >= argCount)
                        {
                            print("error: no count after -mifacount\n");
                            return;
                        }

                        try
                        {
                            stlength = Convert.ToUInt32(args[i + 1]);
                        }
                        catch (Exception)
                        {
                            print("error: -stlength number is ivalid\n");
                            return;
                        }

                        break;

                    case "-stnotfixed":
                        ignoreLength = true;
                        break;

                    case "-o":

                        if (i + 1 >= argCount)
                        {
                            print("error: no destination file after -o\n");
                            return;
                        }

                        destinationFileName = args[i + 1];

                        break;

                    default:

                        break;
                }
            }

            // check

            if (outFileEnabled && destinationFileName == "")
            {
                print("error: no destination file selected but mif or raw output enabled\n");
                return;
            }

            if (sourceFileName == "")
            {
                print("error: no source file\n");
                return;
            }

            // assemble

            assemblerDirector assemblerDirector = new assemblerDirector(debugLog);
            Assembler m_assembler = new Assembler(assemblerDirector);

            hiresCounter hc = new hiresCounter();
            hc.start();


            m_assembler.sourceFileName = sourceFileName;
            m_assembler.destFileName = destinationFileName;

            // load
            print("Load source\n");
            if (m_assembler.loadFile() == false)
            {
                print("error: load file to assemble fail!\n");
                return;
            }

            // set warning level

            m_assembler.warningLevel = warningLevel;

            // assemble

            print("Assemble\n");
            if (m_assembler.assemble() == false)
            {
                return;
            }

            // trace
            if (enableTrace)
            {
                print("Trace opcodes\n");
                m_assembler.trace();
            }

            double dataCount = (double)m_assembler.dataCount;
            double lineCount = (double)m_assembler.lineCount;
            double maxCount = (double)stlength;

            if (ignoreLength)
            {
                maxCount = m_assembler.lastAddress;
            }

            // store
            print("Store to destination\n");
                    
            try
            {
                if (m_assembler.storeFile(m_storeTypes, (uint)maxCount) == false)
                {
                    print("error: store assembled file(s) fail!\n");
                    return;
                }
            }
            catch (Exception ex)
            {
                print("error: store assembled file(s) fail! Reason '" + ex.Message + "'\n");
                return;
            }

            print(string.Format("Use {0} of {1} words ({2:##.##}%)\n", dataCount, maxCount, 100.0 / maxCount * dataCount));
            hc.stop();

            double elapsedTime = hc.elapsed();

            print(string.Format("Finished to assemble {0} lines in {1:.00}s ({2:.0} ls)\n", lineCount, elapsedTime, lineCount / elapsedTime));


            // this is the end

            if (waitKey)
            {
                Console.ReadKey();
            }

        }
        
        static void print(string txt)
        {
            Console.Write(txt);
        }

        public class assemblerDirector : IPrint
        {
            public assemblerDirector(bool debugPrintEnable)
            {
                m_debugPrintEnable = debugPrintEnable;
            }


            public void print(string txt)
            {
                Program.print(txt);
            }

            public void dprint(string txt)
            {
                if (m_debugPrintEnable)
                {
                    Program.print(txt);
                }
            }

            bool m_debugPrintEnable;
        }
        
    }
}
