﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using ucore_assembler_library.IO;
using System.Threading.Tasks;
using System.Text.RegularExpressions;
using System.IO;
using xcore_tools.Common;
using xcore_tools;
using xcore_tools.Data;
using ucore_assembler_library.Data;

namespace ucore_assembler_library
{
    public class Assembler : bitExtraction , IPrint
    {
        public enum storeTypes_t
        {
            st_mif,
            st_raw,
            st_debug
        };

        string m_sourceFileName = "";
        string m_destFilenName = "";
        string m_incDir = "";
        IPrint m_iprint;
        uint m_pc = 0;
        UInt32 m_stopLabelCounter = 0;

        List<IOpcode<UInt16>> m_opcodes;
        List<Line> m_lines;
        Labels m_labels;
        bool m_hazardCheckRegReqLoad;
        bool m_hazardIrqDelaySlot;
        bool m_hazardStall;

        public Assembler(IPrint iprint)
        {
            warningLevel = 2;
            m_iprint = iprint;
            m_hazardCheckRegReqLoad = false;
            m_hazardIrqDelaySlot = false;
            m_hazardStall = false;
            hcoreTrace = false;
        }

        public uint warningLevel { get; set;}

        public int dataCount
        {
            get 
            {
                int totalCount = 0;
                foreach (IOpcode<UInt16> o in m_opcodes)
                {
                    if (o.isBinary)
                    {
                        totalCount+=(int)o.lengthBinary;
                    }
                    else
                    {
                        totalCount++;
                    }
                }
                return totalCount;
            }
        }

        public int lastAddress
        {
            get
            {
                int maxAdr = 0;
                foreach(IOpcode<UInt16> o in m_opcodes)
                {
                    if (o.isBinary)
                    {
                        maxAdr = Math.Max(maxAdr,(int)(o.pc + o.lengthBinary));
                    }
                    else
                    {
                        maxAdr = Math.Max(maxAdr, (int)o.pc + 1);
                    }
                }
                return maxAdr;
            }
        }

        public int lineCount
        {
            get { return m_lines != null ? m_lines.Count : 0; }
        }

        public string sourceFileName
        {
            get {   return m_sourceFileName;    }
            set {   m_sourceFileName = value;   }        
        }

        public string destFileName
        {
            get {   return m_destFilenName;     }
            set {   m_destFilenName = value;    }        
        }

        public void print(string txt)
        {
            m_iprint.print(txt);
        }

        public void dprint(string txt)
        {
            m_iprint.dprint(txt);
        }

        public bool assemble()
        {
            bool result = true;
            dprint("assembler: start\n");

            m_opcodes = new List<IOpcode<UInt16>>();
            m_labels = new Labels();
            result = parse();

            return result;
        }

        public bool loadFile()
        {
            bool result = true;
            dprint("assembler: load file\n");

            LoadText loader = new LoadText(this);

            m_lines = new List<Line>();            
            result = loader.readFile(m_sourceFileName,ref m_lines);
            if (result == true)
            {
                dprint("assembler: " + m_lines.Count() + " Line(s) loaded\n");
            }

            return result;
        }

        public bool storeFile(List<storeTypes_t> stypes,uint ilength)
        {
            string stypesString = " ";
            foreach (storeTypes_t t in stypes)
            {
                stypesString += t.ToString() + " ";
            }
            dprint("assembler: store file(s) mode(" + stypesString + ") length = " + ilength + "\n");

            bool result = true;
            foreach (storeTypes_t t in stypes)
            {
                switch (t)
                {
                    case storeTypes_t.st_mif    : result &= (new SaveMif()).save(m_destFilenName, m_opcodes, ilength); break;
                    case storeTypes_t.st_raw    : result &= (new SaveRaw(this)).save(m_destFilenName, m_opcodes, ilength); break;
                    case storeTypes_t.st_debug  : result &= (new SaveDebug()).save(m_destFilenName, m_opcodes, ilength); break;
                }
            }
            return result;
        }

        public UInt16[] getRawData(uint length)
        {
            UInt16[] rawData = new UInt16[length];
            SortedDictionary<uint, UInt16> wData = new SortedDictionary<uint, UInt16>();

            foreach (IOpcode<UInt16> o in m_opcodes)
            {
                UInt16 value = (UInt16)o.generate();
                uint pc = o.pc;

                if (o.isBinary)
                {
                    for (int i = 0; i < o.lengthBinary; i++)
                    {
                        wData[(uint)(pc + i)] = o.dataBinary[i];
                    }
                }
                else
                {
                    wData[pc] = value;
                }
            }

            for (uint adr = 0; adr < length; adr++)
            {
                UInt16 value = 0xffff;

                if (wData.ContainsKey(adr) == true)
                {
                    value = wData[adr];
                }

                rawData[adr] = value;
            }

            return rawData;
        }

        bool parse()
        {
            bool result = internal_parse();
            if (result == false)
            {
                m_opcodes = null;
            }
            return result;
        }

        const string m_realtime_cmd_value = @"^[\s]+(?<cmd>\.[A-Za-z]+)[\s]+(?<value>[0-9]+[0-9]*)[\s]*(?<comment>;[^\n]*)?$";
        Regex Regexp_realtime_cmd_value = new Regex(m_realtime_cmd_value, RegexOptions.Compiled);

        bool internal_parse()
        {
            //

            dprint("assembler: preprocess\n");

            Preprocessor preProcessor = new Preprocessor(this);
            preProcessor.hcoreTrace = hcoreTrace;
            preProcessor.process(m_lines);

            foreach (KeyValuePair<string, string> kv in preProcessor.defines)
            {
                string prePorcessorDefineName  = kv.Key;
                string preProcessorDefineValue = kv.Value;

                if (preProcessorDefineValue != "")
                {

                    m_labels.insert(prePorcessorDefineName, (uint)Tools.extractInteger(preProcessorDefineValue));
                }
            }
                   
            //

            dprint("assembler: parse/check\n");

            m_pc = 0;

           
            foreach (Line l in m_lines)
            {
                try
                {
                    // assembler commands

                    // check for realtime commands like org, if not found its a normal opcode

                    Match match = Regexp_realtime_cmd_value.Match(l.value);
                    if (match.Success) // its realtime
                    {
                        string cmd = match.Groups["cmd"].Value;
                        string value = match.Groups["value"].Value;

                        if (cmd == ".org")
                        {
                            m_pc = (uint)Tools.extractInteger(value);
                        }
                        else
                        {
                            throw new Exception("illegal command");
                        }
                    }
                    else
                    {
                        // opcode parse

                        Opcode o = new Opcode(this);
                        o.labelDictonary = m_labels;

                        if (o.parse_check(l) == false)
                        {
                            return false;
                        }

                        if (o.isEmpty() == false)   // if not empty
                        {
                            o.pc = m_pc;

                            if (o.isBinary)
                            {
                                m_pc += o.lengthBinary;
                            }
                            else
                            {
                                if (o.onlyLabel() == false) // if not only a label addpc
                                {
                                    m_pc++;
                                }
                            }

                            if (o.setupLabel() == false)
                            {
                                return false;
                            }
                            m_opcodes.Add(o);   // add non empty opcodes
                        }
                    }
                }
                catch (Exception ex)
                {
                    print("Error '" + ex.Message + "' at Line " + l.line + " in File " + l.filename + "\n");
                    return false;
                }
           }

            // 

            dprint("assembler: resolve labels\n");

            bool resolveOk = true;
            // Parallel.ForEach(m_opcodes, o =>
            foreach (IOpcode<UInt16> o in m_opcodes)
            {
                try
                {              
                    if (o.checkResolveImm() == false)
                    {
                        resolveOk = false;
                    }               
                }
                catch (Exception ex)
                {
                    print("Error '" + ex.Message + "' at Line " + ((Opcode)o).line.line + " in File " + ((Opcode)o).line.filename + "\n");
                    return false;
                }
            }

            if (resolveOk == false)
            {
                  return false;
            }

            // check code

            dprint("assembler: check code\n");
            {
                string message;

                List<IOpcode<UInt16>> checkList = new List<IOpcode<UInt16>>();
                for (int i = 0, length = m_opcodes.Count; i < length; i++)
                {
                    Opcode current = (Opcode)m_opcodes[i];
                    if (current.noData() == false)
                    {
                        checkList.Add(current);
                    }
                }

                for (int i = 0, length = checkList.Count; i < length; i++)
                {
                    Opcode current = (Opcode)checkList[i];
                    if (current.checkHazard(warningLevel, checkList, i, out message) == true)
                    {
                        print( message + " detected at Line " + (((Opcode)current).line.line + 1) + " in File " + ((Opcode)current).line.filename + "\n");
                    }  
                }
            }

            return true;
        }
      
        public void trace()
        {
            if (m_opcodes != null)
            {
                foreach (Opcode o in m_opcodes)
                {
                    print(o.trace() + "\n");
                }
            }
        }

        public bool hcoreTrace { get; set; }

        public void traceIncludeStats()
        {
            SortedDictionary<string, int> fileOpcodeCount = new SortedDictionary<string, int>();
            if (m_opcodes != null)
            {
                foreach (Opcode o in m_opcodes)
                {
                    string fileName = o.line.filename;
                    if (fileOpcodeCount.ContainsKey(fileName) == false)
                    {
                        fileOpcodeCount[fileName] = 1;
                    }
                    else
                    {
                        fileOpcodeCount[fileName]++;
                    }
                }
            }

            foreach (KeyValuePair<string, int> kv in fileOpcodeCount)
            {
                print(string.Format("{0,8} opcodes assembled in source file '{1}'\n", kv.Value, kv.Key));
            }
        }

        public void traceOpcodeStats()
        {
            SortedDictionary<string, int> opcodeCount = new SortedDictionary<string, int>();

            if (m_opcodes != null)
            {
                foreach (Opcode o in m_opcodes)
                {
                
                    string opcodeName = o.opcodeName;
                    if (opcodeName != "")
                    {
                        if (opcodeCount.ContainsKey(opcodeName) == false)
                        {
                            opcodeCount[opcodeName] = 1;
                        }
                        else
                        {
                            opcodeCount[opcodeName]++;
                        }
                    }
                }
            }

            foreach (KeyValuePair<string, int> kv in opcodeCount)
            {
                print(string.Format("opcode {0,8} is used {1} times\n",kv.Key,kv.Value));
            }
        }
    }
}
