﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using xcore_tools.Common;
using xcore_tools;
using xcore_tools.Data;

namespace ucore_assembler_library.Data
{
    public class Opcode : bitExtraction, IOpcode<UInt16>
    {
        Labels m_labels = null;
        IPrint m_iprint = null;
        Line m_line = null;
       
        string m_labelName;
        string m_comment;
        paramEncodeType_t m_opcodeParamType;
        string m_opcodeName;
        uint m_reg3;
        uint m_reg2;
        uint m_reg1;
        string m_immText;
        int m_immValue;
        UInt16 m_opcodeData;
        uint m_pc;
        bool m_writeBack;
        bool m_tout;
        bool m_requestData;
        bool m_irqInjectable;
        bool m_canStall;
        bool m_accessIMem;
        UInt16[] m_binary = null;
       
        public Opcode(IPrint iprint)
        {
            m_iprint = iprint;
            m_binary = null;
            Opcode.initPatterns();
            Opcode.initOpcodes();
        }

        public enum hazardCheckType_t
        {
            REG_REQUEST_LOAD,
            IRQ_COMPATIBLE_DELAY_SLOTS,
            STALL_HAZARD
        }

        public bool checkHazard(uint warningLevel, List<IOpcode<UInt16>> data, int index, out string message)
        {
            message = "";
            if (warningLevel >= 0)  // will ever warn
            {
                if ((index + 1) >= data.Count)
                {
                    return false;
                }
                Opcode next = (Opcode)data[index + 1];

                if (m_writeBack == true)    // write back so next could read 
                {
                    // get current destination reg

                    uint destReg = m_reg3 & 7;

                    // get next source regs

                    uint sourceReg1 = 0xff;
                    uint sourceReg2 = 0xff;

                    switch (next.m_opcodeParamType)
                    {
                        case paramEncodeType_t.empty:
                            break;

                        case paramEncodeType_t.dc_imm8:
                            sourceReg1 = (next.m_reg3 & 7);
                            break;

                        case paramEncodeType_t.dc_db_da:
                            sourceReg1 = (next.m_reg2 & 7);
                            sourceReg2 = (next.m_reg1 & 7);
                            break;

                        case paramEncodeType_t.dc_db_imm3:
                            sourceReg2 = (next.m_reg2 & 7);
                            break;

                        case paramEncodeType_t.db_da:
                            sourceReg1 = (next.m_reg3 & 7);
                            sourceReg2 = (next.m_reg2 & 7);
                            break;

                        case paramEncodeType_t.db_imm3:
                            sourceReg2 = (next.m_reg3 & 7);
                            break;

                        case paramEncodeType_t.da:
                            sourceReg2 = (next.m_reg3 & 7);
                            break;

                        case paramEncodeType_t.imm6:
                            break;

                        case paramEncodeType_t.imm11:
                            break;

                        case paramEncodeType_t.imm12:
                            break;

                        case paramEncodeType_t.imm16:
                        case paramEncodeType_t.imm16uc:
                            break;
                    }

                    if (sourceReg1 != 0xff && sourceReg1 == destReg)
                    {
                        message = "Warning: Reg/Reqload hazard! Register r" + destReg + " has write/read hazard";
                        return true;
                    }

                    if (sourceReg2 != 0xff && sourceReg2 == destReg)
                    {
                        message = "Warning: Reg/Reqload hazard! Register r" + destReg + " has write/read hazard";
                        return true;
                    }
                }               
            }
            
            if (warningLevel >= 1)
            {
                if ((index + 1) >= data.Count)
                {
                    return false;
                }
                Opcode nextOpcode = (Opcode)data[index + 1];

                if (m_accessIMem && nextOpcode.m_canStall)
                {
                    message = "Warning: Next instructionhas has stall hazard if access memory from code port";
                    return true;
                }
            }

            if (warningLevel >= 2)
            {
                if (m_irqInjectable == true)
                {
                    if ((index + 4) >= data.Count)
                    {
                        return false;
                    }

                    // check branch in delay slot 
                    for (int i = 1 + index; i <= 4 + index; i++)
                    {
                        Opcode testOpcode = (Opcode)data[i];
                        switch (testOpcode.m_opcodeName)
                        {
                            case "rti":
                            case "brtc":
                            case "brts":
                            case "br":
                            case "jmpi":
                            case "jmpts":
                            case "jmptc":

                                message = "Warning: Instruction (+" + (i - index) + ") is branch in branch delay slot (this is very dangerous if irq is enabled)";
                                return true;

                        }
                    }

                    // check req load

                    for (int i = 2 + index; i <= 4 + index; i++)
                    {
                        Opcode testOpcode = (Opcode)data[i];

                        if (testOpcode.m_requestData == true)
                        {
                            message = "Warning: Instruction (+" + (i - index) + ") request load with no chance to load in this delay slot (request load is not safe in delayslot if irq enabled)";
                            return true;
                        }
                    }
                }
            }

            if (warningLevel >= 3)
            {
                if (m_irqInjectable == true)
                {
                    if ((index + 4) >= data.Count)
                    {
                        return false;
                    }

                    // check req load

                    for (int i = 1 + index; i <= 1 + index; i++)
                    {
                        Opcode testOpcode = (Opcode)data[i];

                        if (testOpcode.m_requestData == true)
                        {
                            message = "Warning: Instruction (+" + (i - index) + ") request load (request load is not safe in delayslot if irq enabled)";
                            return true;
                        }
                    }
                }

                // check t out

                int length = data.Count;
                for (int i = 1 + index; i <= 4 + index; i++)
                {
                    if(i<length)
                    {
                        Opcode testOpcode = (Opcode)data[i];
                        if (testOpcode.m_tout == true)
                        {
                            message = "Warning: Instruction (+" + (i - index) + ") use t as out (t is not safe in delayslot if irq enabled)";
                            return true;
                        }
                    }
                }
            }

            return false;
        }

        public void decode(UInt16 data_in)
        {
            // default

            m_opcodeParamType = paramEncodeType_t.imm16;
            m_opcodeName = "word";

            m_reg3		=0;
			m_reg2		=0;
			m_reg1		=0;
			m_immValue	=0;

            if (extr(data_in, 15) == 0) //	-- reg direct
            {
                m_opcodeParamType = paramEncodeType_t.dc_imm8;
                m_reg3 = extr(data_in, 10, 8);
                m_reg2 = extr(data_in, 10, 8);
                m_reg1 = 0;
                m_immValue = extr(data_in, 7, 0);
               

                switch (extr(data_in, 14, 11))
                {
                    case 0: m_opcodeName = "movei"; break;
                    case 1: m_opcodeName = "moveih"; break;
                    case 2: m_opcodeName = "addi"; break;
                    case 3: m_opcodeName = "subi"; break;
                    case 4: m_opcodeName = "lsri"; break;
                    case 5: m_opcodeName = "asri"; break;
                    case 6: m_opcodeName = "muli"; break;
                    case 7: m_opcodeName = "cmpeqi"; break;
                    case 8: m_opcodeName = "cmploi"; break;
                    case 9: m_opcodeName = "cmplosi"; break;
                    case 10: m_opcodeName = "gpci"; break;
                    case 11: m_opcodeName = "jmpi"; break;
                    case 12: m_opcodeName = "extri"; break;
                    case 13: m_opcodeName = "rqldi"; break;
                    case 14: m_opcodeName = "getsp"; break;
                    case 15: m_opcodeName = "dexti";
                        m_opcodeParamType = paramEncodeType_t.imm11;
                        m_immValue = extr(data_in, 10, 0);
                        break;
                }

               
               
            }
            else
            {
                if (extr(data_in, 14) == 0) //	-- reg, reg, reg / reg, reg, imm3
                {
                    if (extr(data_in, 13) == 0) //	-- reg, reg, reg
                    {
                        switch (extr(data_in, 12, 9))
                        {
                            // reg, reg, reg

                            case 0: m_opcodeName = "add"; break;
                            case 1: m_opcodeName = "sub"; break;
                            case 2: m_opcodeName = "lsr"; break;
                            case 3: m_opcodeName = "asr"; break;
                            case 4: m_opcodeName = "mul"; break;
                            case 5: m_opcodeName = "addt"; break;
                            case 6: m_opcodeName = "subt"; break;
                            case 7: m_opcodeName = "muls"; break;
                            case 8: m_opcodeName = "movets"; break;
                            case 9: m_opcodeName = "and"; break;
                            case 10: m_opcodeName = "or"; break;
                            case 11: m_opcodeName = "xor"; break;
                            case 12: m_opcodeName = "bic"; break;
                            case 13: m_opcodeName = "illegal"; break;
                            case 14: m_opcodeName = "illegal"; break;
                            case 15: m_opcodeName = "dexti"; break;
                        }

                        m_opcodeParamType = paramEncodeType_t.dc_db_da;
                        m_reg3 = extr(data_in, 8, 6);
                        m_reg2 = extr(data_in, 5, 3);
                        m_reg1 = extr(data_in, 2, 0);
                        m_immValue = 0;

                    }
                    else // data_in(13) = '1'	-- reg, reg, imm3
                    {
                        switch (extr(data_in, 12, 9))
                        {
                            // reg, reg, (s) imm3

                            case 0: m_opcodeName = "addqi"; break;
                            case 1: m_opcodeName = "subqi"; break;
                            case 2: m_opcodeName = "lsrqi"; break;
                            case 3: m_opcodeName = "asrqi"; break;
                            case 4: m_opcodeName = "mulqi"; break;
                            case 5: m_opcodeName = "addtqi"; break;
                            case 6: m_opcodeName = "subtqi"; break;
                            case 7: m_opcodeName = "illegal"; break;
                            case 8: m_opcodeName = "edrqldi"; break;
                            case 9: m_opcodeName = "illegal"; break;
                            case 10: m_opcodeName = "illegal"; break;
                            case 11: m_opcodeName = "illegal"; break;
                            case 12: m_opcodeName = "illegal"; break;
                            case 13: m_opcodeName = "illegal"; break;
                            case 14: m_opcodeName = "illegal"; break;
                            case 15: m_opcodeName = "illegal"; break;
                        }

                        m_opcodeParamType = paramEncodeType_t.dc_db_imm3;
                        m_reg3 = extr(data_in, 8, 6);
                        m_reg2 = extr(data_in, 5, 3);
                        m_reg1 = 0;
                        m_immValue = extr(data_in, 2, 0);
                    }
                }
                else		// data(14) = '1'	
                {
                    if (extr(data_in, 13, 12) == 3) // reg, reg / reg, imm3 / reg
                    {
                        if (extr(data_in, 11) == 0) // reg, reg / imm 6
                        {
                            switch (extr(data_in, 10, 6))
                            {
                                case 0x00: m_opcodeName = "cmpeq"; break;
                                case 0x04: m_opcodeName = "cmplo"; break;
                                case 0x08: m_opcodeName = "cmplos"; break;
                                case 0x0c: m_opcodeName = "espage"; break;
                                case 0x10: m_opcodeName = "swp"; break;
                                case 0x14: m_opcodeName = "swptc"; break;
                                case 0x18: m_opcodeName = "st"; break;
                                case 0x1c: m_opcodeName = "stinc"; break;

                                case 0x01: m_opcodeName = "bffo"; break;
                                case 0x05: m_opcodeName = "bset"; break;
                                case 0x09: m_opcodeName = "bclr"; break;
                                case 0x0d: m_opcodeName = "cmple"; break;
                                case 0x11: m_opcodeName = "cmples"; break;
                                case 0x15: m_opcodeName = "rqld"; break;

                                case 0x19: m_opcodeName = "extb"; break;
                                case 0x1d: m_opcodeName = "stwo"; break;

                                case 0x02: m_opcodeName = "andts"; break;
                                case 0x06: m_opcodeName = "orts"; break;
                                case 0x0a: m_opcodeName = "xorts"; break;
                                case 0x0e: m_opcodeName = "bicts"; break;
                                case 0x12: m_opcodeName = "lsrts"; break;
                                case 0x16: m_opcodeName = "asrts"; break;
                                case 0x1a: m_opcodeName = "mults"; break;
                                case 0x1e: m_opcodeName = "mulsts"; break;

                                // imm6

                                case 0x03: m_opcodeName = "erqldi"; break;
                                case 0x07: m_opcodeName = "ssto"; break;
                                case 0x0b: m_opcodeName = "udivinit"; break;
                               
                            }

                            if (((data_in >> 6) & 0x3) == 3) // 	-- imm6
                            {
                                m_opcodeParamType = paramEncodeType_t.imm6;
                                m_reg3 = 0;  
                                m_reg2 = 0;
                                m_reg1 = 0;
                                m_immValue = extr(data_in, 5, 0);
                            }
                            else // reg, reg
                            {
                                m_opcodeParamType = paramEncodeType_t.db_da;
                                m_reg3 = extr(data_in, 5, 3);   // change for disassembler
                                m_reg2 = extr(data_in, 2, 0);
                                m_reg1 = 0;
                                m_immValue = 0;
                            }

                        }
                        else	// data(11) = '1'
                            if (extr(data_in, 10, 4) == ((UInt16)0x7f))	// -- no param
                            {
                                switch (extr(data_in, 3, 0))
                                {
                                    case 0: m_opcodeName = "rqpop"; break;
                                    case 1: m_opcodeName = "cli"; break;
                                    case 2: m_opcodeName = "sei"; break;
                                    case 3: m_opcodeName = "rti"; break;
                                    case 4: m_opcodeName = "tnt"; break;
                                    case 5: m_opcodeName = "vtt"; break;
                                    case 6: m_opcodeName = "exl2derqld"; break;
                                    case 7: m_opcodeName = "epushsadrl"; break;
                                    case 8: m_opcodeName = "epushsadrh"; break;
                                    case 9: m_opcodeName = "epopsadrl"; break;
                                    case 10: m_opcodeName = "epopsadrh"; break;
                                    case 11: m_opcodeName = "udivstep"; break;
                                    case 12: m_opcodeName = "exl2den"; break;       // extension
                                    case 13: m_opcodeName = "exl2denerqld"; break;  // extension
                                    case 14: m_opcodeName = "dump"; break;
                                    case 15: m_opcodeName = "nop"; break;
                                }

                                m_opcodeParamType = paramEncodeType_t.empty;
                                m_reg3 = 0;
                                m_reg2 = 0;
                                m_reg1 = 0;
                                m_immValue = 0;
                            }
                            else
                                if (extr(data_in, 10, 8) == 7) // reg		
                                {
                                    switch (extr(data_in, 7, 3))
                                    {
                                        case 0x00: m_opcodeName = "eld"; break;
                                        case 0x02: m_opcodeName = "push"; break;
                                        case 0x04: m_opcodeName = "gmulhi"; break;
                                        case 0x06: m_opcodeName = "ld_pop"; break;
                                        case 0x08: m_opcodeName = "poparqp"; break;
                                        case 0x0a: m_opcodeName = "neg"; break;
                                        case 0x0c: m_opcodeName = "udivgq"; break;
                                        case 0x0e: m_opcodeName = "udivgr"; break;
                                        case 0x10: m_opcodeName = "udivgqh"; break;
                                        case 0x12: m_opcodeName = "udivgrh"; break;
                                        case 0x14: m_opcodeName = "udivsdd"; break;
                                        case 0x16: m_opcodeName = "udivsdv"; break;
                                        case 0x18: m_opcodeName = "udivsddh"; break;
                                        case 0x1a: m_opcodeName = "udivsdvh"; break;
                                        case 0x1c: m_opcodeName = "erqld"; break;                                       
                                        case 0x01: m_opcodeName = "setsp"; break;
                                        case 0x03: m_opcodeName = "erqldb"; break;
                                        case 0x05: m_opcodeName = "jmpts"; break;
                                        case 0x07: m_opcodeName = "jmptc"; break;
                                        case 0x09: m_opcodeName = "negts"; break;
                                        default: m_opcodeName = "illegal"; break;
                                    }

                                    m_opcodeParamType = paramEncodeType_t.da;
                                    m_reg3 = extr(data_in, 2, 0);
                                    m_reg2 = extr(data_in, 2, 0);
                                    m_reg1 = extr(data_in, 2, 0);
                                    m_immValue = 0;
                                }
                                else	// reg, imm3
                                {
                                    switch (extr(data_in, 10, 8))
                                    {
                                        case 0: m_opcodeName = "illegal"; break;
                                        case 1: m_opcodeName = "illegal"; break;
                                        case 2: m_opcodeName = "illegal"; break;
                                        case 3: m_opcodeName = "est"; break;
                                        case 4: m_opcodeName = "estts"; break;
                                        case 5: m_opcodeName = "getssr"; break;
                                        case 6: m_opcodeName = "setssr"; break;
                                        case 7: m_opcodeName = "illegal"; break;
                                    }

                                    m_opcodeParamType = paramEncodeType_t.db_imm3;
                                    m_reg3 = extr(data_in, 5, 3);
                                    m_reg2 = extr(data_in, 5, 3);
                                    m_reg1 = 0;
                                    m_immValue = extr(data_in, 2, 0);
                                }
                    }
                    else											// imm12
                    {
                        switch (extr(data_in, 13, 12))
                        {
                            case 0: m_opcodeName = "br"; break;
                            case 1: m_opcodeName = "brts"; break;
                            case 2: m_opcodeName = "brtc"; break;
                        }

                        m_opcodeParamType = paramEncodeType_t.imm12;
                        m_reg3 = 0;
                        m_reg2 = 0;
                        m_reg1 = 0;
                        m_immValue = extr(data_in, 11, 0);
                    }
                }
            }

            m_labelName = "";
        }

        public string opcodeName
        {
            get { return m_opcodeName; }
        }

        public ushort generate()
        {
            UInt16 value;

            value = m_opcodeData;

            switch (m_opcodeParamType)
            { 
                case paramEncodeType_t.empty:
                    break;

                case paramEncodeType_t.dc_imm8:
                    value |= (UInt16)((m_reg3 & 7) << 8);                                
                    value |= (UInt16)((m_immValue & 0xff)); 
                    break;
             
                case paramEncodeType_t.dc_db_da:
                    value |= (UInt16)((m_reg3 & 7) << 6); 
                    value |= (UInt16)((m_reg2 & 7) << 3);
                    value |= (UInt16)((m_reg1 & 7));
                    break;

                case paramEncodeType_t.dc_db_imm3:
                    value |= (UInt16)((m_reg3 & 7) << 6);
                    value |= (UInt16)((m_reg2 & 7) << 3);
                    value |= (UInt16)((m_immValue & 0x7));
                    break;

                case paramEncodeType_t.db_da:
                    value |= (UInt16)((m_reg3 & 7) << 3);
                    value |= (UInt16)((m_reg2 & 7));
                    break;

                case paramEncodeType_t.db_imm3:
                    value |= (UInt16)((m_reg3 & 7) << 3);
                    value |= (UInt16)((m_immValue & 0x7));
                    break;

                case paramEncodeType_t.da:
                    value |= (UInt16)((m_reg3 & 7));
                    break;

                case paramEncodeType_t.imm6:
                    value |= (UInt16)((m_immValue & 0x3f));
                    break;

                case paramEncodeType_t.imm11:
                    value |= (UInt16)((m_immValue & 0x7ff));
                    break;

                case paramEncodeType_t.imm12:
                    value |= (UInt16)((m_immValue & 0xfff)); 
                    break;

                case paramEncodeType_t.imm16:
                case paramEncodeType_t.imm16uc:
                    value |= (UInt16)((m_immValue & 0xffff)); 
                    break;
            }
        
            return value;
        }

        public string trace(bool longFormat = true)
        {
            string result = "";

            if (onlyLabel() == true)
            {
                if (longFormat)
                {
                    result += String.Format("{0:X4} {1:X4} [{2}]", m_pc, generate(), m_labelName.PadRight(10));
                }               
            }
            else
            {
                if (longFormat)
                {
                    result += String.Format("{0:X4} {1:X4} [{2}] {3}", m_pc, generate(), m_labelName.PadRight(10), m_opcodeName.PadRight(10));
                }
                else
                {
                    result += String.Format("{0}", m_opcodeName.PadRight(10));
                }
                switch (m_opcodeParamType)
                { 
                     case paramEncodeType_t.empty:
                        break;

                    case paramEncodeType_t.dc_imm8:
                        result += String.Format("r{0},${1:X2} ", m_reg3, ((uint) m_immValue) & 0xff); 
                        break;                 

                    case paramEncodeType_t.dc_db_da:
                        result += String.Format("r{0},r{1},r{2} ", m_reg3, m_reg1, m_reg2);
                        break;

                    case paramEncodeType_t.dc_db_imm3:
                        result += String.Format("r{0},r{1},${2:X1} ", m_reg3, m_reg2, ((uint)m_immValue) & 0x7);
                        break;

                    case paramEncodeType_t.db_da:
                        result += String.Format("r{0},r{1} ", m_reg3, m_reg2);
                        break;

                    case paramEncodeType_t.db_imm3:
                        result += String.Format("r{0},${1:X1} ", m_reg3, ((uint)m_immValue) & 0x7);
                        break;

                    case paramEncodeType_t.da:
                        result += String.Format("r{0} ", m_reg3);
                        break;

                    case paramEncodeType_t.imm6:
                        result += String.Format("${0:X2} ", ((uint)m_immValue) & 0x3f);
                        break;

                    case paramEncodeType_t.imm11:
                        result += String.Format("${0:X3} ", ((uint)m_immValue) & 0x7ff);
                        break;

                    case paramEncodeType_t.imm12:
                        result += String.Format("${0:X3} ", ((uint) m_immValue)  & 0xfff); 
                        break;

                    case paramEncodeType_t.imm16:
                    case paramEncodeType_t.imm16uc:
                        result += String.Format("${0:X4} ", ((uint) m_immValue)  & 0xffff); 
                        break;
                }
                result += m_comment;
            }

            return result;
        }

        public Labels labelDictonary
        {
            get { return m_labels; }
            set { m_labels = value; }
        }
        
        public Line line
        {
            get { return m_line; }
        }

        public uint pc
        {
            get { return m_pc; }
            set { m_pc = value; }
        }

        public bool isBinary
        {
            get { return m_binary != null; }
        }

        public uint lengthBinary
        {
            get
            {
                int length = 0; 
                if (isBinary == true)
                {
                    length = m_binary.Length;   // length is ever in words!                   
                }

                return (uint)length;
            }
        }

        public UInt16[] dataBinary
        {
            get { return m_binary; }
        }

        // parse

        public bool parse_check(Line currentLine)
        {
            if (currentLine.binary) // its binary (from incbin)
            {
                m_binary = currentLine.bData;

                m_immText = "";
                m_line = currentLine;
                m_labelName = "";
                m_comment = "generated by binary (length = " + currentLine.bData.Length + " words)";
                m_opcodeName = "";
            }
            else
            {
                // parse opcode

                Match match = null;
                patternType_t patternType = patternType_t.pt_empty;
                bool patternFound = false;

                foreach (parsePattern<patternType_t> pp in m_parsePatterns)
                {
                    match = pp.regex.Match(currentLine.value);
                    patternType = pp.patternType;

                    if (match.Success == true)
                    {
                        patternFound = true;
                       // m_iprint.dprint("Match at Line " + currentLine.line + " Type " + pp.patternType.ToString() + "\n");
                        break;
                    }
                }

                if (patternFound == false)
                {
                    throw new Exception("Syntax error");
                }

                // is a valid code

                string labelText = match.Groups["label"].Value;
                string commentText = match.Groups["comment"].Value;
                string opcodeText = match.Groups["opcode"].Value;
                string immText = match.Groups["imm"].Value;
                string reg3Text = match.Groups["reg3"].Value;
                string reg2Text = match.Groups["reg2"].Value;
                string reg1Text = match.Groups["reg1"].Value;

                // check opcode and type


                if (patternType != patternType_t.pt_empty)
                {
                    opcodeDefinition o = Opcode.findOpcode(opcodeText, patternType);  // store opcode name/type
                    if (o == null)
                    {
                        throw new Exception("Illegal opcode or parameter");
                    }

                    m_writeBack = o.writeBack;
                    m_tout = o.tout;
                    m_opcodeData = o.opcodeData;
                    m_opcodeParamType = o.paramEncodeType;
                    m_requestData = o.requestData;
                    m_irqInjectable = o.irqInjectable;
                    m_canStall = o.canStall;
                    m_accessIMem = o.accessIMem;
                }

                // store

                m_line = currentLine;
                m_labelName = labelText;
                m_comment = commentText;
                m_opcodeName = opcodeText;

                char[] regTrim = new char[2];
                regTrim[0] = 'r';
                regTrim[1] = 'R';

                try
                {
                    if (reg3Text != "")
                    {
                        reg3Text = reg3Text.TrimStart(regTrim);
                        m_reg3 = Convert.ToUInt32(reg3Text);
                        if (m_reg3 >= 8)
                        {
                            throw new Exception("destination register lower than 0 or greater then 7");
                        }
                    }
                    if (reg2Text != "")
                    {
                        reg2Text = reg2Text.TrimStart(regTrim);
                        m_reg2 = Convert.ToUInt32(reg2Text);
                        if (m_reg2 >= 8)
                        {
                            throw new Exception("source register lower than 0 or greater then 7");
                        }
                    }
                    if (reg1Text != "")
                    {
                        reg1Text = reg1Text.TrimStart(regTrim);
                        m_reg1 = Convert.ToUInt32(reg1Text);
                        if (m_reg1 >= 8)
                        {
                            throw new Exception("source register lower than 0 or greater then 7");
                        }
                    }
                }
                catch (Exception)
                {
                    throw new Exception("Illegal parameter(s)");
                }

                m_immText = immText;
            }
            return true;
        }

        public bool isEmpty()
        {
            return m_opcodeName == "" && m_labelName == "" && m_binary == null;
        }

        public bool onlyLabel()
        {
            return m_opcodeName == "" && m_labelName != "";
        }

        public bool noData()
        {
            return isEmpty() || onlyLabel();
        }

        public bool setupLabel()
        {
            bool result = true;
            if (m_labelName != "")
            {
                if (m_labels.insert(m_labelName, pc) == false)
                {
                    throw new Exception("Label '" + m_labelName + "' already exists");
                }
            }

            return result;
        }

        static Regex m_immRegex = new Regex(@"^([>]*|[3>]*|[6>]*)[A-Za-z]+[A-Za-z0-9_]+$", RegexOptions.Compiled);

        public bool checkResolveImm()
        {
             if (m_immText != "")
             {
                 Match match = m_immRegex.Match(m_immText);

                 bool wasLabel = false;
                 if (match.Success == true)
                 {
                     wasLabel = true;
                     int shiftRight = 0;

                     while (m_immText[0] == '3' && m_immText[1] == '>')    // 3 bit shift
                     {
                         m_immText = m_immText.Remove(0, 2);
                         shiftRight += 3;
                     }

                     while (m_immText[0] == '6' && m_immText[1] == '>')    // 6 bit shift
                     {
                         m_immText = m_immText.Remove(0, 2);
                         shiftRight += 6;
                     }

                     while (m_immText[0] == '>')   // right shift 8 bit
                     {
                        m_immText = m_immText.Remove(0, 1);
                        shiftRight += 8;
                     }
                     

                     uint address;
                     label_TypeEnum labelType;
                     if (m_labels.search(m_immText, out address, out labelType) == false)
                     {
                         throw new Exception("Cant resolve Label \"" + m_immText + "\" ");
                     }

                     m_immValue = (int)address >> shiftRight;
                     m_iprint.dprint("Resolve Label '" + m_immText + "' to " + String.Format("${0:X4}", m_immValue) + "\n");
                 }
                 else
                 {
                     // value
                     try
                     {
                         m_immValue = Tools.extractInteger(m_immText);
                     }
                     catch (Exception)
                     {
                         throw new Exception("Illegal immediate value");
                     }
                 }

                 // check bounds

                 if (   m_opcodeParamType == paramEncodeType_t.dc_db_imm3   ||
                        m_opcodeParamType == paramEncodeType_t.db_imm3 )
                 {
                     Immediate.translate(m_immValue, 3, false, wasLabel);                   
                 }                
                 else if (  m_opcodeParamType == paramEncodeType_t.dc_imm8 )
                 {
                     Immediate.translate(m_immValue, 8, false, wasLabel);          
                 }
                 else if (m_opcodeParamType == paramEncodeType_t.imm6)
                 {
                     Immediate.translate(m_immValue, 6, true, wasLabel);                    
                 }
                 else if (m_opcodeParamType == paramEncodeType_t.imm11)
                 {
                     Immediate.translate(m_immValue, 11, false, wasLabel);                      
                 }
                 else if (m_opcodeParamType == paramEncodeType_t.imm12)
                 {
                     if (wasLabel == true)
                     {
                         m_immValue = m_immValue - ((int)m_pc) - 4;
                     }

                     Immediate.translate(m_immValue, 12, true, false);                  
                 }
                 else if (m_opcodeParamType == paramEncodeType_t.imm16)
                 {
                     if (m_immValue < -32768 || m_immValue > 65535)
                     {
                         throw new Exception("Illegal immediate value unsigned (0..65535) / signed (-32768..32767)");
                     }

                     m_immValue &= 0xffff; // if signed cut to 16bit
                 }
                 else if (m_opcodeParamType == paramEncodeType_t.imm16uc)
                 {
                     m_immValue &= 0xffff; // if signed cut to 16bit
                 } 
             }

            return true;
        }


        // parse patterns

        //
        // p1 (label) (comment)
        // p2 (label)  opcode  (comment)
        // p3 (label)  opcode (rX) (comment)
        // p4 (label)  opcode (rX,rY) (comment)
        // p5 (label)  opcode (rX,rY,rZ) (comment)
        // p6 (label)  opcode (rX,rY,imm) (comment)
        // p7 (label)  opcode (rX,imm) (comment)
        // p8 (label)  opcode (imm) (comment)
        // 

        const string p1 = @"^(?<label>[A-Za-z]+[A-Za-z0-9_]+)?[\s]*(?<comment>;[^\n]*)?$";
        const string p2 = @"^(?<label>[A-Za-z]+[A-Za-z0-9_]+)?[\s]+(?<opcode>[A-Za-z_2]+)[\s]*(?<comment>;[^\n]*)?$";
        const string p3 = @"^(?<label>[A-Za-z]+[A-Za-z0-9_]+)?[\s]+(?<opcode>[A-Za-z_2]+)[\s]+(?<reg3>[rR][0-9]+)[\s]*(?<comment>;[^\n]*)?$";
        const string p4 = @"^(?<label>[A-Za-z]+[A-Za-z0-9_]+)?[\s]+(?<opcode>[A-Za-z_2]+)[\s]+(?<reg3>[rR][0-9]+)[\s]*,[\s]*(?<reg2>[rR][0-9]+)[\s]*(?<comment>;[^\n]*)?$";
        const string p5 = @"^(?<label>[A-Za-z]+[A-Za-z0-9_]+)?[\s]+(?<opcode>[A-Za-z_2]+)[\s]+(?<reg3>[rR][0-9]+)[\s]*,[\s]*(?<reg2>[rR][0-9]+)[\s]*,[\s]*(?<reg1>[rR][0-9]+)[\s]*(?<comment>;[^\n]*)?$";
        const string p6 = @"^(?<label>[A-Za-z]+[A-Za-z0-9_]+)?[\s]+(?<opcode>[A-Za-z_2]+)[\s]+(?<reg3>[rR][0-9]+)[\s]*,[\s]*(?<reg2>[rR][0-9]+)[\s]*,[\s]*(?<imm>[\>\+\-%$A-Za-z0-9]+)[\s]*(?<comment>;[^\n]*)?$";
        const string p7 = @"^(?<label>[A-Za-z]+[A-Za-z0-9_]+)?[\s]+(?<opcode>[A-Za-z_2]+)[\s]+(?<reg3>[rR][0-9]+)[\s]*,[\s]*(?<imm>[\'\>\+\-%$A-Za-z0-9]+)[\s]*(?<comment>;[^\n]*)?$";
        const string p8 = @"^(?<label>[A-Za-z]+[A-Za-z0-9_]+)?[\s]+(?<opcode>[A-Za-z_2]+)[\s]+(?<imm>[\'\>\+\-%$A-Za-z0-9]+)[\s]*(?<comment>;[^\n]*)?$";

        public enum patternType_t
        {
            pt_empty,
            pt_opcode,
            pt_opcode_reg_imm,
            pt_opcode_reg,
            pt_opcode_reg_reg,
            pt_opcode_reg_reg_reg,
            pt_opcode_reg_reg_imm,
            pt_opcode_imm
        }

        public enum paramEncodeType_t
        {
            empty,
            dc_imm8,
            dc_db_da,
            dc_db_imm3,
            db_da,
            imm6,
            db_imm3,
            da,
            imm11,
            imm12,
            imm16,
            imm16uc
        }

        static bool isParsePatterParamCompatible(patternType_t pat,paramEncodeType_t par)
        {
            switch (pat)
            {
                case patternType_t.pt_empty:                        
                    break;

                case patternType_t.pt_opcode:
                    if(par == paramEncodeType_t.empty)
                    {
                        return true;
                    }
                    break;

                 case patternType_t.pt_opcode_reg_imm:
                    if( par == paramEncodeType_t.dc_imm8 ||
                        par == paramEncodeType_t.db_imm3)
                    {
                        return true;
                    }
                    break;

                case patternType_t.pt_opcode_reg:
                    if( par == paramEncodeType_t.da)
                    {
                        return true;
                    }
                    break;

                case patternType_t.pt_opcode_reg_reg:
                    if( par == paramEncodeType_t.db_da)
                    {
                        return true;
                    }
                    break;

                 case patternType_t.pt_opcode_reg_reg_reg:
                    if( par == paramEncodeType_t.dc_db_da)
                    {
                        return true;
                    }
                    break;

                case patternType_t.pt_opcode_reg_reg_imm:
                    if( par == paramEncodeType_t.dc_db_imm3)
                    {
                        return true;
                    }
                    break;

                case patternType_t.pt_opcode_imm:
                    if (par == paramEncodeType_t.imm11 || 
                        par == paramEncodeType_t.imm12 ||
                        par == paramEncodeType_t.imm16 ||
                        par == paramEncodeType_t.imm16uc ||
                        par == paramEncodeType_t.imm6 )
                    {
                        return true;
                    }
                    break;
            }

            return false;
        }



        static List<parsePattern<patternType_t>> m_parsePatterns = null;

        static void initPatterns()
        {
            if (m_parsePatterns == null)
            {
                m_parsePatterns = new List<parsePattern<patternType_t>>();
                m_parsePatterns.Add(new parsePattern<patternType_t>(new Regex(p1, RegexOptions.Compiled), patternType_t.pt_empty));
                m_parsePatterns.Add(new parsePattern<patternType_t>(new Regex(p2, RegexOptions.Compiled), patternType_t.pt_opcode));
                m_parsePatterns.Add(new parsePattern<patternType_t>(new Regex(p3, RegexOptions.Compiled), patternType_t.pt_opcode_reg));
                m_parsePatterns.Add(new parsePattern<patternType_t>(new Regex(p4, RegexOptions.Compiled), patternType_t.pt_opcode_reg_reg));
                m_parsePatterns.Add(new parsePattern<patternType_t>(new Regex(p5, RegexOptions.Compiled), patternType_t.pt_opcode_reg_reg_reg));
                m_parsePatterns.Add(new parsePattern<patternType_t>(new Regex(p6, RegexOptions.Compiled), patternType_t.pt_opcode_reg_reg_imm));
                m_parsePatterns.Add(new parsePattern<patternType_t>(new Regex(p7, RegexOptions.Compiled), patternType_t.pt_opcode_reg_imm));
                m_parsePatterns.Add(new parsePattern<patternType_t>(new Regex(p8, RegexOptions.Compiled), patternType_t.pt_opcode_imm));
            }
        }

        
        // opcodes available    

        public class opcodeDefinition
        {
            public opcodeDefinition(string iname, paramEncodeType_t iparamEncodeType, UInt16 iopcodeData, bool iwriteBack = false, bool itout = false,bool irequestData = false, bool iirqInjectable = false, bool icanStall = false, bool iaccessIMem = false)
            {
                name = iname;
                opcodeData = iopcodeData;
                paramEncodeType = iparamEncodeType;
                writeBack = iwriteBack;
                tout = itout;
                requestData = irequestData;
                irqInjectable = iirqInjectable;
                canStall = icanStall;
                accessIMem = iaccessIMem;
            }

            public paramEncodeType_t paramEncodeType
            {
                get { return m_paramEncodeType; }
                set { m_paramEncodeType = value; }
            }

            public UInt16 opcodeData
            {
                get { return m_opcodeData; }
                set { m_opcodeData = value; }
            }

            public string name
            {
                get { return m_name; }
                set { m_name = value; }
            }

            public bool writeBack
            {
                get { return m_writeBack; }
                set { m_writeBack = value; }
            }

            public bool tout
            {
                get { return m_tout; }
                set { m_tout = value; }
            }

            public bool requestData
            {
                get { return m_requestData; }
                set { m_requestData = value; }
            }

            public bool irqInjectable
            {
                get { return m_irqInjectable; }
                set { m_irqInjectable = value; }
            }

            public bool canStall
            {
                get { return m_canStall; }
                set { m_canStall = value; }
            }

            public bool accessIMem
            {
                get { return m_accessIMem; }
                set { m_accessIMem = value; }
            }

            paramEncodeType_t m_paramEncodeType;
            UInt16 m_opcodeData;
            string m_name;
            bool m_writeBack;
            bool m_tout;
            bool m_requestData;
            bool m_irqInjectable;
            bool m_canStall;
            bool m_accessIMem;
        }

        static SortedDictionary<string, List<opcodeDefinition>> m_opcodes = null;

        public List<opcodeDefinition> getOpcodes()
        {
            List<opcodeDefinition> rList = new List<opcodeDefinition>();
            foreach (KeyValuePair<string, List<opcodeDefinition>> kv in m_opcodes)
            {
                if (kv.Value != null)
                {
                    foreach (opcodeDefinition oD in kv.Value)
                    {
                        rList.Add(oD);
                    }
                }
            }

            return rList;
        }

        static void initOpcodes()
        {
            if (m_opcodes == null)
            {
                //

                string opcodeName;
                List<opcodeDefinition> t;

                // reg imm8

                m_opcodes = new SortedDictionary<string, List<opcodeDefinition>>();

                opcodeName = "movei";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.dc_imm8, 0x0000,true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "moveih";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.dc_imm8, 0x0800, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "addi";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.dc_imm8, 0x1000, true,true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "subi";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.dc_imm8, 0x1800, true,true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "lsri";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.dc_imm8, 0x2000, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "asri";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.dc_imm8, 0x2800, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "muli";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.dc_imm8, 0x3000, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "cmpeqi";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.dc_imm8, 0x3800,false,true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "cmploi";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.dc_imm8, 0x4000, false, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "cmplosi";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.dc_imm8, 0x4800, false, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "gpci";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.dc_imm8, 0x5000, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "jmpi";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.dc_imm8, 0x5800,false,false,false,true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "extri";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.dc_imm8, 0x6000, false, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "rqldi";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.dc_imm8, 0x6800,false,false,true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "getsp";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.dc_imm8, 0x7000, true, false, false, false, false));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "dexti";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.imm11, 0x7800));
                m_opcodes.Add(opcodeName, t);    

                // reg reg reg

                opcodeName = "add";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.dc_db_da, 0x8000, true,true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "sub";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.dc_db_da, 0x8200, true,true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "lsr";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.dc_db_da, 0x8400, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "asr";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.dc_db_da, 0x8600, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "mul";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.dc_db_da, 0x8800, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "addt";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.dc_db_da, 0x8a00, true, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "subt";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.dc_db_da, 0x8c00, true, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "muls";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.dc_db_da, 0x8e00, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "movets";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.dc_db_da, 0x9000,true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "and";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.dc_db_da, 0x9200, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "or";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.dc_db_da, 0x9400, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "xor";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.dc_db_da, 0x9600, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "bic";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.dc_db_da, 0x9800, true));
                m_opcodes.Add(opcodeName, t);

                // reg reg imm3

                opcodeName = "addqi";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.dc_db_imm3, 0xa000, true, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "subqi";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.dc_db_imm3, 0xa200, true, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "lsrqi";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.dc_db_imm3, 0xa400, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "asrqi";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.dc_db_imm3, 0xa600, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "mulqi";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.dc_db_imm3, 0xa800, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "addtqi";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.dc_db_imm3, 0xaa00, true, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "subtqi";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.dc_db_imm3, 0xac00, true, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "edrqldi";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.dc_db_imm3, 0xb000,false,false,false,false,true,false));
                m_opcodes.Add(opcodeName, t);

                // imm12

                opcodeName = "br";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.imm12, 0xc000, false, false, false, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "brts";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.imm12, 0xd000, false, false, false, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "brtc";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.imm12, 0xe000, false, false, false, true));
                m_opcodes.Add(opcodeName, t);


                // reg reg

                opcodeName = "cmpeq";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.db_da, 0xf000, false, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "cmplo";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.db_da, 0xf100, false, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "cmplos";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.db_da, 0xf200, false, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "esadr";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.db_da, 0xf300));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "swp";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.db_da, 0xf400, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "swptc";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.db_da, 0xf500, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "st";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.db_da, 0xf600));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "stinc";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.db_da, 0xf700,true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "bffo";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.db_da, 0xf040, true,true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "bset";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.db_da, 0xf140, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "bclr";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.db_da, 0xf240, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "cmple";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.db_da, 0xf340, false, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "cmples";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.db_da, 0xf440, false, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "rqld";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.db_da, 0xf540,false,false,true));
                m_opcodes.Add(opcodeName, t);
            
                opcodeName = "extb";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.db_da, 0xf640,true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "stwo";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.db_da, 0xf740));
                m_opcodes.Add(opcodeName, t);

                //

                opcodeName = "andts";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.db_da, 0xf080, true));
                m_opcodes.Add(opcodeName, t);

                 opcodeName = "orts";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.db_da, 0xf180, true));
                m_opcodes.Add(opcodeName, t);

                 opcodeName = "xorts";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.db_da, 0xf280, true));
                m_opcodes.Add(opcodeName, t);

                  opcodeName = "bicts";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.db_da, 0xf380, true));
                m_opcodes.Add(opcodeName, t);

                  opcodeName = "lsrts";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.db_da, 0xf480, true));
                m_opcodes.Add(opcodeName, t);

                  opcodeName = "asrts";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.db_da, 0xf580, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "mults";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.db_da, 0xf680, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "mulsts";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.db_da, 0xf780, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "erqldi";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.imm6, 0xf0c0, false, false, false, false, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "ssto";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.imm6, 0xf1c0));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "udivinit";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.imm6, 0xf2c0));
                m_opcodes.Add(opcodeName, t);
             
                // reg imm3

                opcodeName = "est";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.db_imm3, 0xfb00, false, false, false, false, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "estts";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.db_imm3, 0xfc00, false, false, false, false, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "getssr";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.db_imm3, 0xfd00,true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "setssr";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.db_imm3, 0xfe00));
                m_opcodes.Add(opcodeName, t);       

                // reg

                opcodeName = "eld";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.da, 0xff00, true, false, false, false, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "push";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.da, 0xff10, false, false, false, false, false, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "gmulhi";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.da, 0xff20, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "ld";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.da, 0xff30, true, false, false, false, false, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "pop";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.da, 0xff30, true, false, false, false, false, true));  // equal to ld
                m_opcodes.Add(opcodeName, t);

                opcodeName = "poparqp";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.da, 0xff40, true, false, false, false, false, true));  // equal to ld
                m_opcodes.Add(opcodeName, t);

                opcodeName = "neg";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.da, 0xff50,true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "udivgq";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.da, 0xff60,true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "udivgr";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.da, 0xff70,true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "udivgqh";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.da, 0xff80,true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "udivgrh";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.da, 0xff90,true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "udivsdd";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.da, 0xffa0));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "udivsdv";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.da, 0xffb0));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "udivsddh";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.da, 0xffc0));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "udivsdvh";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.da, 0xffd0));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "erqld";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.da, 0xffe0, false, false, false, false, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "setsp";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.da, 0xff08));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "erqldb";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.da, 0xff18, false, false, false, false, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "jmpts";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.da, 0xff28, false, false, false, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "jmptc";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.da, 0xff38, false, false, false, true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "negts";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.da, 0xff48, true));
                m_opcodes.Add(opcodeName, t);

                // no param
          
                opcodeName = "rqpop";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.empty, 0xfff0));
                m_opcodes.Add(opcodeName, t);
             
                opcodeName = "cli";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.empty, 0xfff1));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "sei";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.empty, 0xfff2));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "rti";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.empty, 0xfff3));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "tnt";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.empty, 0xfff4,false,true));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "vtt";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.empty, 0xfff5, false, true));
                m_opcodes.Add(opcodeName, t);
            
                opcodeName = "exl2derqld";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.empty, 0xfff6));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "epushsadrl";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.empty, 0xfff7));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "epushsadrh";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.empty, 0xfff8));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "epopsadrl";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.empty, 0xfff9));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "epopsadrh";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.empty, 0xfffa));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "udivstep";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.empty, 0xfffb));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "exl2den";            // extension
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.empty, 0xfffc));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "exl2denerqld";            // extension
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.empty, 0xfffd));
                m_opcodes.Add(opcodeName, t);
        
                opcodeName = "dump";    // only emulator instruction to dump data
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.empty, 0xfffe));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "nop";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.empty, 0xffff));
                m_opcodes.Add(opcodeName, t);
                     
                // pseudo

                opcodeName = "word";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.imm16, 0x0000));
                m_opcodes.Add(opcodeName, t);

                opcodeName = "worduc";
                t = new List<opcodeDefinition>();
                t.Add(new opcodeDefinition(opcodeName, paramEncodeType_t.imm16uc, 0x0000));
                m_opcodes.Add(opcodeName, t);
            }
        }

        static opcodeDefinition findOpcode(string name,patternType_t type)
        {          
            if (m_opcodes != null)
            {
                if (m_opcodes.ContainsKey(name))
                {
                    List<opcodeDefinition> opcodes = m_opcodes[name];
                    foreach (opcodeDefinition o in opcodes)
                    {
                        if (isParsePatterParamCompatible(type, o.paramEncodeType) == true)
                        {
                            return o;
                        }
                    }
                   
                }
            }

            return null;
        }
    }
}
