﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using ucore_assembler_library;
using xcore_tools.Common;
using xcore_tools;

namespace ucore_emulator.core.components.memory.ctrl
{

    public class Audio : bitExtraction, IPrint, IClocked 
    {
        IPrint m_iprint;
        bool m_dacFifoIrq;
        Queue<UInt64> m_dacFifo = new Queue<UInt64>();
        UInt64 m_cycleCount = 0;
        UInt64 m_cycleTrigger = 1109;

        const int mc_audioOutMaskMax = 0xffff;
        int m_audioOutMask = 0x7f;
        int m_audioOutCount = 0;
        int[,] m_audioOut = new int[2, mc_audioOutMaskMax + 1];

        int m_fifoMaxUsedW;
        int m_fifoMinUsedW;

        hiresCounter m_hiresCounter;
        double m_currentMaxFrequency;

        //

        bool m_dacEnabled;

        // real audio

        AudioRAWOut m_audioRaw;
        int m_sampleFrequency = 2000; // default 4kHz, because emulator is to slow
        public int sampleFrequency { get { return m_sampleFrequency; } }

        //
        
        public Audio(IPrint iprint)
        {
            enableRealAudio = false;
            m_iprint = iprint;
            m_currentMaxFrequency = 0;
            m_hiresCounter = new hiresCounter();
            m_hiresCounter.start();
            m_audioRaw = new AudioRAWOut(sampleFrequency);

            reset();
        }

        public bool enableRealAudio { get; set; }

        public bool dacFifoIrq
        {
            get { return m_dacFifoIrq; }
        }
 
        // read to debug/view

        public double currentMaxFrequency
        {
            get { return m_currentMaxFrequency; }
        }

        public void clearDacFifo()
        {
            m_dacFifo.Clear();
            m_fifoMaxUsedW = -1;
            m_fifoMinUsedW = 99999999;
        }

        public int fifoMaxUsedW
        {
            get { return m_fifoMaxUsedW; }
        }

        public int fifoMinUsedW
        {
            get { return m_fifoMinUsedW; }
        }

        public int audioMask
        {
            get 
            {
                return (int)Math.Log((double)m_audioOutMask+1,2);
            }

            set 
            {
                if (value <= 16 && value > 2)
                {
                    m_audioOutMask = (1 << value)-1;
                }
            }
            
        }

        public int[,] audioOut 
        {
            get { return m_audioOut; }
        }

        public int audioOutLength
        {
            get { return m_audioOutMask + 1; }
        }

        public bool isDacEnabled
        {
            get { return m_dacEnabled; }
        }

        public int fifoUseCount
        {
            get { return m_dacFifo.Count; }
        }

        //

        public void reset()
        {
            m_dacFifoIrq = false;
            m_dacFifo.Clear();
            m_cycleCount = 0;
            m_dacEnabled = false;

            for (int i = 0; i <= mc_audioOutMaskMax; i++)   // total clear of buffer
            {
                m_audioOut[0,i] = 0;
                m_audioOut[1,i] = 0;
            }

            m_fifoMaxUsedW = -1;
            m_fifoMinUsedW = 99999999;
        }

        public void step()
        {
            int fifoUsedW = m_dacFifo.Count;

            m_fifoMaxUsedW = fifoUsedW > m_fifoMaxUsedW ? fifoUsedW : m_fifoMaxUsedW;
            m_fifoMinUsedW = fifoUsedW < m_fifoMinUsedW ? fifoUsedW : m_fifoMinUsedW;

            if(m_cycleCount == m_cycleTrigger)
            {
                m_cycleCount = 0;

                if (m_dacFifo.Count != 0)
                {
                    // performance measurment

                    m_hiresCounter.stop();
                    m_currentMaxFrequency = 1.0 / m_hiresCounter.elapsed();
                    m_hiresCounter.start();

                    //

                    UInt64 dacWord = m_dacFifo.Dequeue();

                    UInt32 leftRaw  = (UInt32) extr(dacWord, 63, 40);
                    UInt32 rightRaw = (UInt32) extr(dacWord, 39, 16);

                    int left = exts(leftRaw, 23);
                    int right = exts(leftRaw, 23);

                    switch (extr(dacWord, 15, 8))
                    {
                        case 1: m_dacEnabled = (dacWord & 1) == 1; break;
                        default: break;
                    }

                    if (m_dacEnabled)
                    {
                        // real audio

                        if (enableRealAudio)
                        {
                            m_audioRaw.pushLRData(left, right);
                        }

                        // debug / view

                        m_audioOut[0, m_audioOutCount] = left;
                        m_audioOut[1, m_audioOutCount] = right;

                        m_audioOutCount++;
                        m_audioOutCount &=m_audioOutMask;
                    }
                }
            }
            else
            {
                m_cycleCount ++;
            }

            m_dacFifoIrq = m_dacFifo.Count < 128;          
        }

        //

        public void pushDacFifo(UInt64 data)
        {
            m_dacFifo.Enqueue(data);
        }

        //

        #region IPrint
        public void print(string txt)
        {
            if (m_iprint != null)
            {
                m_iprint.print(txt);
            }
        }

        public void dprint(string txt)
        {
            if (m_iprint != null)
            {
                m_iprint.dprint(txt);
            }
        }
        #endregion

    }
}
