﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using NAudio.Wave;

namespace ucore_emulator.core.components.memory.ctrl
{
    public abstract class WaveProvider32 : IWaveProvider
    {
        private WaveFormat waveFormat;

        public WaveProvider32()
            : this(44100, 1)
        {
        }

        public WaveProvider32(int sampleRate, int channels)
        {
            SetWaveFormat(sampleRate, channels);
        }

        public void SetWaveFormat(int sampleRate, int channels)
        {
            this.waveFormat = WaveFormat.CreateIeeeFloatWaveFormat(sampleRate, channels);
        }

        public int Read(byte[] buffer, int offset, int count)
        {
            WaveBuffer waveBuffer = new WaveBuffer(buffer);
            int samplesRequired = count / 4;
            int samplesRead = Read(waveBuffer.FloatBuffer, offset / 4, samplesRequired);
            return samplesRead * 4;
        }

        public abstract int Read(float[] buffer, int offset, int sampleCount);

        public WaveFormat WaveFormat
        {
            get { return waveFormat; }
        }
    }

    public class SampleWaveProvider32 : WaveProvider32
    {
        
        int m_sampleFrequency;
        float m_masterVolume; 
        Int64 m_bigCounter;

        object _lock = new object();

        Queue<float> m_rightQueue = new Queue<float>();
        Queue<float> m_leftQueue = new Queue<float>();

        float right = 0f;
        float left = 0f;     // left channel is track 1 & 2

        public SampleWaveProvider32(int sampleFrequency)
        {
            m_sampleFrequency = sampleFrequency;
            m_masterVolume = 1.0f;
        }

        public void setMasterVolume(float masterVolume)
        {
            m_masterVolume = masterVolume;
        }

        public void pushLRData(float left, float right)
        {
            lock (_lock)
            {
                m_leftQueue.Enqueue(left);
                m_rightQueue.Enqueue(right);
            }
        }

        public override int Read(float[] buffer, int offset, int sampleCount)
        {           
            int sampleRate = WaveFormat.SampleRate;
            for (int n = 0; n < sampleCount; n += 2)
            {
                right = left = 0;

                lock (_lock)
                {

                    if (m_rightQueue.Count >= m_sampleFrequency)
                    {
                        right = m_rightQueue.Dequeue();
                    }

                    if (m_leftQueue.Count >= m_sampleFrequency)
                    {
                        left = m_leftQueue.Dequeue();
                    }
                }

                buffer[n + offset] = right * m_masterVolume;
                buffer[n + offset + 1] = left * m_masterVolume;
               
                m_bigCounter++;
            }
            return sampleCount;
        }
    }

    public class AudioRAWOut
    {   
        SampleWaveProvider32 m_sampleWaveProvider;

        private WaveOut waveOut;

        public AudioRAWOut(int sampleFrequency)
        {           
            initAudio(sampleFrequency);
        }
     
        public void setMasterVolume(float masterVolume)
        {
            m_sampleWaveProvider.setMasterVolume(masterVolume);
        }

        public void pushLRData(int left, int right)
        {
            float l = left;
            float r = right;

            l /= 65536f;
            r /= 65536f;

            m_sampleWaveProvider.pushLRData(l, r);
        }

        public void cleanup()
        {
            waveOut.Stop();
            waveOut.Dispose();
            waveOut = null;
        }
       
        private void initAudio(int sampleFrequency)
        {
            m_sampleWaveProvider = new SampleWaveProvider32(sampleFrequency);
            m_sampleWaveProvider.SetWaveFormat(sampleFrequency, 2); // ?? Hz Stero
            waveOut = new WaveOut();
            waveOut.DesiredLatency = 300;   // 300ms latency (increase if you have sound problems)
            waveOut.Init(m_sampleWaveProvider);

            waveOut.Play();
        }
    }
}
