library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity utility_ctrl_ps2 is
		port 
		(
				-- 
				
				clk				: in std_logic;
				reset				: in std_logic;
				
				-- data out
				
				data_out			: out std_logic_vector(7 downto 0);
				data_out_ready : out std_logic;
				data_out_ack	: in std_logic;
				
				-- data in
				
				data_in					: in std_logic_vector(7 downto 0);
				data_in_we				: in std_logic;
				data_in_ready			: out std_logic;
				data_in_wrs_finish	: out std_logic;
				
				--
				
				ps2_clk_inout		: inout std_logic;
				ps2_dat_inout		: inout std_logic
									
		);
end entity;

architecture rtl of utility_ctrl_ps2 is

-- parity gen

function odd_par(arg : std_logic_vector(7 downto 0) )
return std_logic is
	variable odd : std_logic;
begin
	odd := '0';
	
	for i in 7 downto 0 loop
		odd := odd xor arg(i);
	end loop;

	return odd;
end;
	
--

component sc_fifo IS
	
	GENERIC 
	(
		FIFO_DEPTH 		: natural;
		FIFO_WIDTH 		: natural;
		FIFO_SHOWAHEAD : string;
		FIFO_USERAM		: string
		
	);
	PORT
	(
		clock		: IN STD_LOGIC ;
		data		: IN STD_LOGIC_VECTOR (FIFO_WIDTH-1 DOWNTO 0);
		rdreq		: IN STD_LOGIC ;
		sclr		: IN STD_LOGIC ;
		wrreq		: IN STD_LOGIC ;
		empty		: OUT STD_LOGIC ;
		full		: OUT STD_LOGIC ;
		q		: OUT STD_LOGIC_VECTOR (FIFO_WIDTH-1 DOWNTO 0)
	);
end component;	
	
-- send fifo
	
signal ps2_output_rdreq		  : std_logic := '0';
signal ps2_output_fifo_empty : std_logic := '0';
signal ps2_output_fifo_full  : std_logic := '0';		
signal ps2_output_q  : STD_LOGIC_VECTOR (7 DOWNTO 0) := (others => '0');	
	
-- store fifo	
	
signal ps2_input_fifo_empty : std_logic := '0';
signal ps2_input_fifo_full  : std_logic := '0';	
	
--	
	
type ps2_in_clk_event_t is 
	( 
		ICLK_RISE,
		ICLK_FALL,
		ICLK_NO_CHANGE
	);	
	
signal internal_ps2_clk : std_logic	 := '0';
signal internal_ps2_data : std_logic := '0';
signal ps2_in_clk_event : ps2_in_clk_event_t	:= ICLK_NO_CHANGE;


--

signal read_frame_state_counter	: integer range 0 to 10 := 0;		
signal frame 							: std_logic_vector(7 downto 0) := (others => '0');
signal frame_recived					: std_logic := '0';
signal frame_read_in_progress		: std_logic := '0';
signal frame_write_in_progress	: std_logic := '0';

--
	
begin

	-----------------------------------------------------------------------------------------------------------
	-- data to ps2 fifo
	-----------------------------------------------------------------------------------------------------------
	
	ps2_output_fifo : sc_fifo
	generic map 
	(
		FIFO_DEPTH 		=> 4,		-- max 4 to send
		FIFO_WIDTH 		=> 8,
		FIFO_SHOWAHEAD => "ON",
		FIFO_USERAM		=> "OFF" -- use les
	)
	port map 
	(
		clk,
		data_in,
		ps2_output_rdreq,
		reset,
		data_in_we,
		ps2_output_fifo_empty,
		ps2_output_fifo_full,
		ps2_output_q		
	);	
	
	data_in_ready 			<= not ps2_output_fifo_full;
	data_in_wrs_finish	<= ps2_output_fifo_empty;
	
	-----------------------------------------------------------------------------------------------------------
	-- data from ps2 fifo
	-----------------------------------------------------------------------------------------------------------

	ps2_input_fifo :sc_fifo
	generic map 
	(
		FIFO_DEPTH 		=> 16,	-- max 16 to recive
		FIFO_WIDTH 		=> 8,
		FIFO_SHOWAHEAD => "ON",
		FIFO_USERAM		=> "OFF" -- use les
	)
	port map 
	(
		clk,
		frame,
		data_out_ack,
		reset,
		frame_recived,
		ps2_input_fifo_empty,
		ps2_input_fifo_full,
		data_out		
	);	
	
	data_out_ready <= not ps2_input_fifo_empty;

	-----------------------------------------------------------------------------------------------------------
	-- data buffer
	-----------------------------------------------------------------------------------------------------------
	
	process(clk,ps2_clk_inout,ps2_dat_inout)
	begin
		if rising_edge(clk) then
			internal_ps2_clk 	<= ps2_clk_inout;
			internal_ps2_data <= ps2_dat_inout;
		end if;
	end process;

	-----------------------------------------------------------------------------------------------------------
	-- ps2 clk scan
	-----------------------------------------------------------------------------------------------------------
	
	process(clk)
		variable last_internal_ps2_clk : std_logic := '1';
		variable tmp : std_logic_vector(1 downto 0) := (others => '0');
	begin
		if rising_edge(clk) then
		
			tmp := internal_ps2_clk & last_internal_ps2_clk;
		
			case tmp is
				when "00" 	=> ps2_in_clk_event <= ICLK_NO_CHANGE;
				when "01" 	=> ps2_in_clk_event <= ICLK_FALL;
				when "10" 	=> ps2_in_clk_event <= ICLK_RISE;
				when "11" 	=> ps2_in_clk_event <= ICLK_NO_CHANGE;
				when others => ps2_in_clk_event <= ICLK_NO_CHANGE;
			end case;
										
			last_internal_ps2_clk := internal_ps2_clk;
		end if;
	end process;
	
	--------------------------------------------------------------------------------------------------------------
	-- ps read data
	--------------------------------------------------------------------------------------------------------------
	
	-- state
	
	process(clk)		
	begin
		if rising_edge(clk) then
			if reset = '1' then
				read_frame_state_counter <= 0;	
			else
				if ps2_in_clk_event = ICLK_FALL then	-- ever on falling internal_ps2_clk edege							
					case read_frame_state_counter is
						when 0 					=> if internal_ps2_data = '0' and frame_write_in_progress = '0' then		-- start 
															read_frame_state_counter <= 1;							
														end if;								 
						when 1 to 8 			=> read_frame_state_counter <= read_frame_state_counter + 1;								
						when 9 					=> read_frame_state_counter <= 10;								
						when 10 					=>	read_frame_state_counter <= 0;					
					end case;									
				end if;
			end if;
		end if;
	end process;
	
	-----------------------------------------------------------------------------------------------------------
	-- data
	-----------------------------------------------------------------------------------------------------------
	
	process(clk)		
		variable tmp_frame_valid			: std_logic := '0';
	begin
		if rising_edge(clk) then
		
			if reset = '1' then
				frame_recived	<= '0';
				frame_read_in_progress <= '0';	
			else
				-- default
				
				frame_recived	<= '0';

				if ps2_in_clk_event = ICLK_FALL then	-- ever on falling internal_ps2_clk edege
					case read_frame_state_counter is
						when 0 					=> if internal_ps2_data = '0' and frame_write_in_progress = '0' then		-- start bit								
															frame_read_in_progress <= '1';	-- read is now in progress
														end if;
									 
						when 1 to 8 			=> frame <= internal_ps2_data & frame(7 downto 1);
						
						when 9 					=> if odd_par(frame) = internal_ps2_data then -- parity ok
															tmp_frame_valid := '0';
														else
															tmp_frame_valid := '1';
														end if;
														
						when 10 					=>	if internal_ps2_data = '1' then		-- stop bit -- data ok
															if tmp_frame_valid = '1' and ps2_input_fifo_full = '0' then										
																frame_recived <= '1';
																frame_read_in_progress <= '0';	-- read is now finished
															end if;									
														end if;																				
					end case;									
				end if;
			end if;
		end if;
	end process;
		
	--------------------------------------------------------------------------------------------------------------
	-- ps write data
	--------------------------------------------------------------------------------------------------------------
	
	process(clk)
		variable tmp_frame_state_counter	: integer range 0 to 5 := 0;
		variable tmp_frame_bit_counter	: integer range 0 to 10 := 0;
		variable tmp_clk_down_counter		: integer range 0 to 1065 := 0;
		variable tmp_frame_data : std_logic_vector(7 downto 0) := (others => '0');
	begin
		if rising_edge(clk) then
		
			-- default
		
			ps2_clk_inout 		<= 'Z';
			ps2_dat_inout 		<= 'Z';
			ps2_output_rdreq 	<= '0';
					
			case tmp_frame_state_counter is		
					when 0 => 	
									frame_write_in_progress <= '0';
									
									if ps2_output_fifo_empty = '0' and frame_read_in_progress = '0' then
										tmp_frame_data 			:= ps2_output_q;
										ps2_output_rdreq			<= '1';
										frame_write_in_progress <= '1';
										
										tmp_frame_state_counter := 1;
										tmp_clk_down_counter		:= 1065;
										
									end if;
					when 1 =>	-- set clk at minum 100us low
									-- clk frequency is 106.481481481 MHz
									
									if tmp_clk_down_counter = 0 then
										tmp_frame_state_counter := 2;
										tmp_frame_bit_counter   := 0;
									else
										tmp_clk_down_counter := tmp_clk_down_counter - 1;
									end if;
									
									ps2_clk_inout <= '0';	-- clk pull down
									
					when 2 =>	ps2_dat_inout <= '0';
									if ps2_in_clk_event = ICLK_FALL then
										tmp_frame_state_counter := 3;
									end if;
												
					when 3 =>	ps2_dat_inout <= tmp_frame_data(tmp_frame_bit_counter);	--
					
									if ps2_in_clk_event = ICLK_FALL then
										tmp_frame_bit_counter := tmp_frame_bit_counter + 1;
										if tmp_frame_bit_counter = 8 then 
											tmp_frame_state_counter := 4;
										else
											tmp_frame_state_counter := 3;											
										end if;
									end if;
									
									
									
					when 4 =>	ps2_dat_inout <= not odd_par(data_in);
									if ps2_in_clk_event = ICLK_FALL then
										tmp_frame_state_counter := 5;
									end if;
									
					when 5 =>	if ps2_in_clk_event = ICLK_FALL then	-- get ack
										if internal_ps2_data = '1' then
											-- stop by ps2
										end if;
										tmp_frame_state_counter := 0;
									end if;
																				
			end case;
		end if;
	end process;
	
end rtl;
